@@ -1,24 +1,24 @@
 /*
  *   stunnel       TLS offloading and load-balancing proxy
- *   Copyright (C) 1998-2015 Michal Trojnara <Michal.Trojnara@mirt.net>
+ *   Copyright (C) 1998-2022 Michal Trojnara <Michal.Trojnara@stunnel.org>
  *
  *   This program is free software; you can redistribute it and/or modify it
  *   under the terms of the GNU General Public License as published by the
  *   Free Software Foundation; either version 2 of the License, or (at your
  *   option) any later version.
- * 
+ *
  *   This program is distributed in the hope that it will be useful,
  *   but WITHOUT ANY WARRANTY; without even the implied warranty of
  *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  *   See the GNU General Public License for more details.
- * 
+ *
  *   You should have received a copy of the GNU General Public License along
  *   with this program; if not, see <http://www.gnu.org/licenses>.
- * 
+ *
  *   Linking stunnel statically or dynamically with other modules is making
  *   a combined work based on stunnel. Thus, the terms and conditions of
  *   the GNU General Public License cover the whole combination.
- * 
+ *
  *   In addition, as a special exception, the copyright holder of stunnel
  *   gives you permission to combine stunnel with free software programs or
  *   libraries that are released under the GNU LGPL and with code included
@@ -26,7 +26,7 @@
  *   modified versions of such code, with unchanged license). You may copy
  *   and distribute such a system following the terms of the GNU GPL for
  *   stunnel and the licenses of the other code concerned.
- * 
+ *
  *   Note that people who make modified versions of stunnel are not obligated
  *   to grant this special exception for their modified versions; it is their
  *   choice whether to do so. The GNU General Public License gives permission
@@ -40,15 +40,31 @@
 
 #include "common.h"
 
+#if defined(USE_PTHREAD) || defined(USE_WIN32)
+#define USE_OS_THREADS
+#endif
+
+#ifdef USE_OS_THREADS
+#ifndef USE_WIN32
+    /* FIXME: waiting for a single terminate_pipe socket
+     * from multiple threads causes a deadlock on WIN32 */
+#define USE_TERMINATE_PIPE
+#endif /* USE_WIN32 */
+#endif /* USE_OS_THREADS */
+
 /**************************************** forward declarations */
 
 typedef struct tls_data_struct TLS_DATA;
+typedef struct sock_opt_struct SOCK_OPT;
 
 /**************************************** data structures */
 
-/* non-zero constants for the "redirect" option */
-#define REDIRECT_ON         1
-#define REDIRECT_OFF        2
+#ifdef USE_PTHREAD
+    typedef pthread_t THREAD_ID;
+#endif
+#ifdef USE_WIN32
+    typedef HANDLE THREAD_ID;
+#endif
 
 #if defined (USE_WIN32)
 #define ICON_IMAGE HICON
@@ -64,7 +80,7 @@ typedef enum {
 } ICON_TYPE;
 
 typedef enum {
-    LOG_MODE_NONE,
+    LOG_MODE_BUFFER,
     LOG_MODE_ERROR,
     LOG_MODE_INFO,
     LOG_MODE_CONFIGURED
@@ -73,7 +89,8 @@ typedef enum {
 typedef enum {
     LOG_ID_SEQUENTIAL,
     LOG_ID_UNIQUE,
-    LOG_ID_THREAD
+    LOG_ID_THREAD,
+    LOG_ID_PROCESS
 } LOG_ID;
 
 typedef enum {
@@ -99,8 +116,9 @@ typedef struct name_list_struct {
 } NAME_LIST;
 
 typedef struct sockaddr_list {                          /* list of addresses */
-    SOCKADDR_UNION *addr;                           /* the list of addresses */
-    unsigned *rr_ptr, rr_val;             /* current address for round-robin */
+    struct sockaddr_list *parent;   /* used by copies to locate their parent */
+    SOCKADDR_UNION *addr;                     /* array of resolved addresses */
+    unsigned start;              /* initial address for round-robin failover */
     unsigned num;                             /* how many addresses are used */
     int passive;                                         /* listening socket */
     NAME_LIST *names;                          /* a list of unresolved names */
@@ -108,12 +126,15 @@ typedef struct sockaddr_list {                          /* list of addresses */
 
 #ifndef OPENSSL_NO_COMP
 typedef enum {
-    COMP_NONE, COMP_DEFLATE, COMP_ZLIB
+    COMP_NONE,                           /* empty compression algorithms set */
+    COMP_DEFLATE,            /* default OpenSSL's compression algorithms set */
+    COMP_ZLIB,          /* additional historic ZLIB compression algorithm id */
+    STUNNEL_COMPS                   /* number of compression algorithms sets */
 } COMP_TYPE;
 #endif /* !defined(OPENSSL_NO_COMP) */
 
 typedef struct {
-        /* some data for SSL initialization in ssl.c */
+        /* some data for TLS initialization in ssl.c */
 #ifndef OPENSSL_NO_COMP
     COMP_TYPE compression;                               /* compression type */
 #endif /* !defined(OPENSSL_NO_COMP) */
@@ -126,7 +147,6 @@ typedef struct {
 #ifdef HAVE_CHROOT
     char *chroot_dir;
 #endif
-    unsigned long dpid;
     char *pidfile;
 #endif
 
@@ -137,7 +157,7 @@ typedef struct {
     char *output_file;
     FILE_MODE log_file_mode;
 
-        /* user interface configuraion */
+        /* user interface configuration */
 #ifdef ICON_IMAGE
     ICON_IMAGE icon[ICON_NONE];                  /* user-specified GUI icons */
 #endif
@@ -166,9 +186,9 @@ typedef struct servername_list_struct SERVERNAME_LIST;/* forward declaration */
 
 #ifndef OPENSSL_NO_PSK
 typedef struct psk_keys_struct {
-    char *identity;
+    const char *identity;                  /* the OpenSSL API requires const */
     unsigned char *key_val;
-    size_t key_len;
+    unsigned key_len;
     struct psk_keys_struct *next;
 } PSK_KEYS;
 typedef struct psk_table_struct {
@@ -177,16 +197,25 @@ typedef struct psk_table_struct {
 } PSK_TABLE;
 #endif /* !defined(OPENSSL_NO_PSK) */
 
+#if OPENSSL_VERSION_NUMBER>=0x10000000L
+typedef struct ticket_key_struct {
+    unsigned char *key_val;
+    int key_len;
+} TICKET_KEY;
+#endif /* OpenSSL 1.0.0 or later */
+
 typedef struct service_options_struct {
     struct service_options_struct *next;   /* next node in the services list */
-    SSL_CTX *ctx;                                            /*  SSL context */
+    SSL_CTX *ctx;                                            /*  TLS context */
     char *servname;        /* service name for logging & permission checking */
+    int ref;                   /* reference counter for delayed deallocation */
 
         /* service-specific data for stunnel.c */
 #ifndef USE_WIN32
     uid_t uid;
     gid_t gid;
 #endif
+    int bound_ports;                /* number of ports bound to this service */
 
         /* service-specific data for log.c */
     int log_level;                                /* debug level for logging */
@@ -197,12 +226,14 @@ typedef struct service_options_struct {
     size_t stack_size;                         /* stack size for this thread */
 #endif
 
+        /* some global data for network.c */
+    SOCK_OPT *sock_opts;
+
         /* service-specific data for verify.c */
     char *ca_dir;                              /* directory for hashed certs */
     char *ca_file;                       /* file containing bunches of certs */
     char *crl_dir;                              /* directory for hashed CRLs */
     char *crl_file;                       /* file containing bunches of CRLs */
-    int verify_level;
 #ifndef OPENSSL_NO_OCSP
     char *ocsp_url;
     unsigned long ocsp_flags;
@@ -214,14 +245,24 @@ typedef struct service_options_struct {
 
         /* service-specific data for ctx.c */
     char *cipher_list;
+#ifndef OPENSSL_NO_TLS1_3
+    char *ciphersuites;
+#endif /* TLS 1.3 */
     char *cert;                                             /* cert filename */
     char *key;                               /* pem (priv key/cert) filename */
     long session_size, session_timeout;
-    long ssl_options_set;
+#if OPENSSL_VERSION_NUMBER>=0x10100000L
+    int security_level;
+#endif /* OpenSSL 1.1.0 or later */
+    uint64_t ssl_options_set;
 #if OPENSSL_VERSION_NUMBER>=0x009080dfL
-    long ssl_options_clear;
+    uint64_t ssl_options_clear;
 #endif /* OpenSSL 0.9.8m or later */
+#if OPENSSL_VERSION_NUMBER>=0x10100000L
+    int min_proto_version, max_proto_version;
+#else /* OPENSSL_VERSION_NUMBER<0x10100000L */
     SSL_METHOD *client_method, *server_method;
+#endif /* OPENSSL_VERSION_NUMBER<0x10100000L */
     SOCKADDR_UNION sessiond_addr;
 #ifndef OPENSSL_NO_TLSEXT
     char *sni;
@@ -233,32 +274,39 @@ typedef struct service_options_struct {
     PSK_TABLE psk_sorted;
 #endif /* !defined(OPENSSL_NO_PSK) */
 #ifndef OPENSSL_NO_ECDH
-    int curve;
+    char *curves;
 #endif /* !defined(OPENSSL_NO_ECDH) */
 #ifndef OPENSSL_NO_ENGINE
     ENGINE *engine;                        /* engine to read the private key */
 #endif /* !defined(OPENSSL_NO_ENGINE) */
+#if OPENSSL_VERSION_NUMBER>=0x10000000L
+    TICKET_KEY *ticket_key;              /* key for handling session tickets */
+    TICKET_KEY *ticket_mac;            /* key for protecting session tickets */
+#endif /* OpenSSL 1.0.0 or later */
 
         /* service-specific data for client.c */
-    SOCKET fd;     /* file descriptor accepting connections for this service */
-    SSL_SESSION *session;                           /* recently used session */
     char *exec_name;                          /* program name for local mode */
 #ifdef USE_WIN32
     char *exec_args;                     /* program arguments for local mode */
 #else
     char **exec_args;                    /* program arguments for local mode */
 #endif
-    SOCKADDR_UNION local_addr, source_addr;
-    SOCKADDR_LIST connect_addr, redirect_addr;
+    SOCKADDR_UNION source_addr;
+    SOCKADDR_LIST local_addr, connect_addr, redirect_addr;
+    SOCKET *local_fd;                 /* array of accepting file descriptors */
+    SSL_SESSION **connect_session;   /* per-destination client session cache */
+    SSL_SESSION *session;    /* previous client session for delayed resolver */
     int timeout_busy;                       /* maximum waiting for data time */
     int timeout_close;                          /* maximum close_notify time */
     int timeout_connect;                           /* maximum connect() time */
     int timeout_idle;                        /* maximum idle connection time */
     enum {FAILOVER_RR, FAILOVER_PRIO} failover;         /* failover strategy */
+    unsigned rr;   /* per-service sequential number for round-robin failover */
     char *username;
 
         /* service-specific data for protocol.c */
-    char * protocol;
+    char *protocol;
+    NAME_LIST *protocol_header;
     char *protocol_host;
     char *protocol_domain;
     char *protocol_username;
@@ -274,6 +322,10 @@ typedef struct service_options_struct {
 
         /* on/off switches */
     struct {
+        unsigned request_cert:1;        /* request a peer certificate */
+        unsigned require_cert:1;        /* require a client certificate */
+        unsigned verify_chain:1;        /* verify certificate chain */
+        unsigned verify_peer:1;         /* verify peer certificate */
         unsigned accept:1;              /* endpoint: accept */
         unsigned client:1;
         unsigned delayed_lookup:1;
@@ -282,10 +334,8 @@ typedef struct service_options_struct {
 #endif
         unsigned local:1;               /* outgoing interface specified */
         unsigned retry:1;               /* loop remote+program */
+        unsigned session_resume:1;      /* enable session resumption */
         unsigned sessiond:1;
-#ifndef OPENSSL_NO_TLSEXT
-        unsigned sni:1;                 /* endpoint: sni */
-#endif /* !defined(OPENSSL_NO_TLSEXT) */
 #ifndef USE_WIN32
         unsigned pty:1;
         unsigned transparent_src:1;
@@ -300,8 +350,11 @@ typedef struct service_options_struct {
         unsigned nonce:1;               /* send and verify OCSP nonce */
 #endif /* !defined(OPENSSL_NO_OCSP) */
 #ifndef OPENSSL_NO_DH
-        unsigned dh_needed:1;
+        unsigned dh_temp_params:1;
 #endif /* OPENSSL_NO_DH */
+#ifndef USE_WIN32
+        unsigned log_stderr:1;          /* a copy of the global switch */
+#endif /* USE_WIN32 */
     } option;
 } SERVICE_OPTIONS;
 
@@ -327,13 +380,13 @@ typedef union {
     struct timeval timeval_val;
 } OPT_UNION;
 
-typedef struct {
-    char *opt_str;
+struct sock_opt_struct {
+    const char *opt_str;
     int  opt_level;
     int  opt_name;
     VAL_TYPE opt_type;
     OPT_UNION *opt_val[3];
-} SOCK_OPT;
+};
 
 typedef enum {
     CONF_RELOAD, CONF_FILE, CONF_FD
@@ -353,6 +406,7 @@ typedef struct {
     unsigned allocated;
 #endif
 #endif
+    int main_thread;
 } s_poll_set;
 
 typedef struct disk_file {
@@ -361,7 +415,7 @@ typedef struct disk_file {
 #else
     int fd;
 #endif
-    /* the inteface is prepared to easily implement buffering if needed */
+    /* the interface is prepared to easily implement buffering if needed */
 } DISK_FILE;
 
     /* definitions for client.c */
@@ -377,83 +431,98 @@ typedef enum {
     RENEG_DETECTED /* renegotiation detected */
 } RENEG_STATE;
 
-typedef struct {
-    jmp_buf err; /* exception handler needs to be 16-byte aligned on Itanium */
-    SSL *ssl; /* SSL connnection */
+typedef struct client_data_struct {
+    jmp_buf *exception_pointer;
+
+    SSL *ssl;                                              /* TLS connection */
     SERVICE_OPTIONS *opt;
     TLS_DATA *tls;
 
-    SOCKADDR_UNION peer_addr; /* peer address */
+#ifdef USE_OS_THREADS
+    THREAD_ID thread_id;
+#endif
+#ifndef USE_FORK
+    struct client_data_struct *thread_prev, *thread_next;
+#endif
+
+    SOCKADDR_UNION peer_addr;                                /* peer address */
     socklen_t peer_addr_len;
-    SOCKADDR_UNION *bind_addr; /* address to bind() the socket */
-    SOCKADDR_LIST connect_addr; /* either copied or resolved dynamically */
-    FD local_rfd, local_wfd; /* read and write local descriptors */
-    FD remote_fd; /* remote file descriptor */
-        /* IP for explicit local bind or transparent proxy */
-    unsigned long pid; /* PID of the local process */
-    SOCKET fd; /* temporary file descriptor */
-    RENEG_STATE reneg_state; /* used to track renegotiation attempts */
-    unsigned long long seq; /* sequential thread number for logging */
+    char *accepted_address;    /* textual representation of the peer address */
+    SOCKADDR_UNION *bind_addr;               /* address to bind() the socket */
+    SOCKADDR_LIST connect_addr;     /* either copied or resolved dynamically */
+    unsigned idx;              /* actually connected address in connect_addr */
+    FD local_rfd, local_wfd;             /* read and write local descriptors */
+    FD remote_fd;                                  /* remote file descriptor */
+    unsigned long pid;                           /* PID of the local process */
+    SOCKET fd;                                  /* temporary file descriptor */
+    RENEG_STATE reneg_state;         /* used to track renegotiation attempts */
+    unsigned long long seq;          /* sequential thread number for logging */
+    unsigned rr;    /* per-client sequential number for round-robin failover */
 
     /* data for transfer() function */
-    char sock_buff[BUFFSIZE]; /* socket read buffer */
-    char ssl_buff[BUFFSIZE]; /* SSL read buffer */
-    size_t sock_ptr, ssl_ptr; /* index of the first unused byte */
-    FD *sock_rfd, *sock_wfd; /* read and write socket descriptors */
-    FD *ssl_rfd, *ssl_wfd; /* read and write SSL descriptors */
-    uint64_t sock_bytes, ssl_bytes; /* bytes written to socket and SSL */
-    s_poll_set *fds; /* file descriptors */
-    uintptr_t redirect; /* redirect to another destination after failed auth */
+    char sock_buff[BUFFSIZE];                          /* socket read buffer */
+    char ssl_buff[BUFFSIZE];                              /* TLS read buffer */
+    size_t sock_ptr, ssl_ptr;              /* index of the first unused byte */
+    FD *sock_rfd, *sock_wfd;            /* read and write socket descriptors */
+    FD *ssl_rfd, *ssl_wfd;                 /* read and write TLS descriptors */
+    uint64_t sock_bytes, ssl_bytes;       /* bytes written to socket and TLS */
+    s_poll_set *fds;                                     /* file descriptors */
+    struct {
+        unsigned psk:1;                            /* PSK identity was found */
+    } flag;
 } CLI;
 
 /**************************************** prototypes for stunnel.c */
 
 #ifndef USE_FORK
-extern long max_clients;
-extern volatile long num_clients;
+extern int max_clients;
+extern int num_clients;
 #endif
+extern SOCKET signal_pipe[2];
+extern SOCKET terminate_pipe[2];
 
 void main_init(void);
 int main_configure(char *, char *);
 void main_cleanup(void);
 int drop_privileges(int);
 void daemon_loop(void);
-void unbind_ports(void);
-int bind_ports(void);
-void signal_post(int);
+void signal_post(uint8_t);
 #if !defined(USE_WIN32) && !defined(USE_OS2)
-void child_status(void);  /* dead libwrap or 'exec' process detected */
+void pid_status_hang(const char *);
 #endif
 void stunnel_info(int);
 
 /**************************************** prototypes for options.c */
 
-extern char *configuration_file;
+extern unsigned number_of_sections;
 
 int options_cmdline(char *, char *);
 int options_parse(CONF_TYPE);
 void options_defaults(void);
 void options_apply(void);
+void options_free(int);
+
+void service_up_ref(SERVICE_OPTIONS *);
+void service_free(SERVICE_OPTIONS *);
 
 /**************************************** prototypes for fd.c */
 
 #ifndef USE_FORK
 void get_limits(void); /* setup global max_clients and max_fds */
 #endif
-SOCKET s_socket(int, int, int, int, char *);
-int s_pipe(int[2], int, char *);
-int s_socketpair(int, int, int, SOCKET[2], int, char *);
-SOCKET s_accept(SOCKET, struct sockaddr *, socklen_t *, int, char *);
+SOCKET s_socket(int, int, int, int, const char *);
+int s_pipe(int[2], int, const char *);
+int s_socketpair(int, int, int, SOCKET[2], int, const char *);
+SOCKET s_accept(SOCKET, struct sockaddr *, socklen_t *, int, const char *);
 void set_nonblock(SOCKET, unsigned long);
 
 /**************************************** prototypes for log.c */
 
-#if !defined(USE_WIN32) && !defined(__vms)
-void syslog_open(void);
-void syslog_close(void);
-#endif
-int log_open(void);
-void log_close(void);
+#define SINK_SYSLOG 1
+#define SINK_OUTFILE 2
+
+int log_open(int);
+void log_close(int);
 void log_flush(LOG_MODE);
 void s_log(int, const char *, ...)
 #ifdef __GNUC__
@@ -461,13 +530,15 @@ void s_log(int, const char *, ...)
 #else
     ;
 #endif
+void s_vlog(int, const char *, va_list);
 char *log_id(CLI *);
-void fatal_debug(char *, const char *, int);
+void fatal_debug(const char *, const char *, int) NORETURN;
 #define fatal(a) fatal_debug((a), __FILE__, __LINE__)
 void ioerror(const char *);
 void sockerror(const char *);
 void log_error(int, int, const char *);
 char *s_strerror(int);
+void bin2hexstring(const unsigned char *, size_t, char *, size_t);
 
 /**************************************** prototypes for pty.c */
 
@@ -479,25 +550,30 @@ DH *get_dh2048(void);
 
 /**************************************** prototypes for cron.c */
 
+#ifdef USE_OS_THREADS
+extern THREAD_ID cron_thread_id;
+#endif
+
 int cron_init(void);
 
 /**************************************** prototypes for ssl.c */
 
-extern int index_cli, index_opt, index_redirect, index_addr;
+extern int index_ssl_cli, index_ssl_ctx_opt;
+extern int index_session_authenticated, index_session_connect_address;
 
+int fips_default(void);
+int fips_available(void);
+void crypto_init(char *);
 int ssl_init(void);
 int ssl_configure(GLOBAL_OPTIONS *);
 
 /**************************************** prototypes for ctx.c */
 
-typedef struct {
-    SERVICE_OPTIONS *section;
-    char pass[PEM_BUFSIZE];
-} UI_DATA;
+extern SERVICE_OPTIONS *current_section;
 
 #ifndef OPENSSL_NO_DH
 extern DH *dh_params;
-extern int dh_needed;
+extern int dh_temp_params;
 #endif /* OPENSSL_NO_DH */
 
 int context_init(SERVICE_OPTIONS *);
@@ -505,51 +581,69 @@ int context_init(SERVICE_OPTIONS *);
 void psk_sort(PSK_TABLE *, PSK_KEYS *);
 PSK_KEYS *psk_find(const PSK_TABLE *, const char *);
 #endif /* !defined(OPENSSL_NO_PSK) */
-void sslerror(char *);
+#ifndef OPENSSL_NO_ENGINE
+UI_METHOD *ui_stunnel(void);
+#endif /* !defined(OPENSSL_NO_ENGINE) */
+void print_session_id(SSL_SESSION *);
+void sslerror(const char *);
 
 /**************************************** prototypes for verify.c */
 
 int verify_init(SERVICE_OPTIONS *);
+void print_client_CA_list(const STACK_OF(X509_NAME) *);
 char *X509_NAME2text(X509_NAME *);
 
 /**************************************** prototypes for network.c */
 
 s_poll_set *s_poll_alloc(void);
 void s_poll_free(s_poll_set *);
-void s_poll_init(s_poll_set *);
+void s_poll_init(s_poll_set *, int);
 void s_poll_add(s_poll_set *, SOCKET, int, int);
 void s_poll_remove(s_poll_set *, SOCKET);
 int s_poll_canread(s_poll_set *, SOCKET);
 int s_poll_canwrite(s_poll_set *, SOCKET);
 int s_poll_hup(s_poll_set *, SOCKET);
 int s_poll_rdhup(s_poll_set *, SOCKET);
+int s_poll_err(s_poll_set *, SOCKET);
 int s_poll_wait(s_poll_set *, int, int);
 void s_poll_dump(s_poll_set *, int);
+void s_poll_sleep(int, int);
 
 #ifdef USE_WIN32
-#define SIGNAL_RELOAD_CONFIG    1
-#define SIGNAL_REOPEN_LOG       2
-#define SIGNAL_TERMINATE        3
+#define SIGNAL_TERMINATE        1
+#define SIGNAL_RELOAD_CONFIG    2
+#define SIGNAL_REOPEN_LOG       3
+#define SIGNAL_CONNECTIONS      4
 #else
+#define SIGNAL_TERMINATE        SIGTERM
 #define SIGNAL_RELOAD_CONFIG    SIGHUP
 #define SIGNAL_REOPEN_LOG       SIGUSR1
-#define SIGNAL_TERMINATE        SIGTERM
+#define SIGNAL_CONNECTIONS      SIGUSR2
 #endif
 
-int set_socket_options(SOCKET, int);
+int socket_options_set(SERVICE_OPTIONS *, SOCKET, int);
 int make_sockets(SOCKET[2]);
 int original_dst(const SOCKET, SOCKADDR_UNION *);
 
 /**************************************** prototypes for client.c */
 
 CLI *alloc_client_session(SERVICE_OPTIONS *, SOCKET, SOCKET);
-void *client_thread(void *);
+#if defined(USE_WIN32) || defined(USE_OS2)
+unsigned __stdcall
+#else
+void *
+#endif
+    client_thread(void *);
 void client_main(CLI *);
+void client_free(CLI *);
+void throw_exception(CLI *, int) NORETURN;
 
 /**************************************** prototypes for network.c */
 
+int get_socket_error(const SOCKET);
 int s_connect(CLI *, SOCKADDR_UNION *, socklen_t);
 void s_write(CLI *, SOCKET fd, const void *, size_t);
+size_t s_read_eof(CLI *, SOCKET fd, void *, size_t);
 void s_read(CLI *, SOCKET fd, void *, size_t);
 void fd_putline(CLI *, SOCKET, const char *);
 char *fd_getline(CLI *, SOCKET);
@@ -561,10 +655,17 @@ void fd_printf(CLI *, SOCKET, const char *, ...)
     ;
 #endif
 void s_ssl_write(CLI *, const void *, int);
+size_t s_ssl_read_eof(CLI *, void *, int);
 void s_ssl_read(CLI *, void *, int);
 char *ssl_getstring(CLI *c);
 char *ssl_getline(CLI *c);
 void ssl_putline(CLI *c, const char *);
+void ssl_printf(CLI *, const char *, ...)
+#ifdef __GNUC__
+    __attribute__((format(printf, 2, 3)));
+#else
+    ;
+#endif
 
 /**************************************** prototype for protocol.c */
 
@@ -575,17 +676,23 @@ typedef enum {
     PROTOCOL_LATE
 } PHASE;
 
-char *protocol(CLI *, SERVICE_OPTIONS *opt, const PHASE);
+#ifdef USE_WIN32
+extern HWND capwin_hwnd;
+extern LONG capwin_connectivity;
+#endif
+
+const char *protocol(CLI *, SERVICE_OPTIONS *opt, const PHASE);
 
 /**************************************** prototypes for resolver.c */
 
-void resolver_init();
+void resolver_init(void);
+int use_ipv6(void);
 
 unsigned name2addr(SOCKADDR_UNION *, char *, int);
-unsigned hostport2addr(SOCKADDR_UNION *, char *, char *, int);
+unsigned hostport2addr(SOCKADDR_UNION *, char *, const char *, int);
 
 unsigned name2addrlist(SOCKADDR_LIST *, char *);
-unsigned hostport2addrlist(SOCKADDR_LIST *, char *, char *);
+unsigned hostport2addrlist(SOCKADDR_LIST *, char *, const char *);
 
 void addrlist_clear(SOCKADDR_LIST *, int);
 unsigned addrlist_dup(SOCKADDR_LIST *, const SOCKADDR_LIST *);
@@ -605,21 +712,8 @@ const char *s_gai_strerror(int);
 #endif
 
 #ifdef USE_WIN32
-
 /* rename some locally shadowed declarations */
 #define getnameinfo     local_getnameinfo
-
-#ifndef _WIN32_WCE
-typedef int (CALLBACK * GETADDRINFO) (const char *,
-    const char *, const struct addrinfo *, struct addrinfo **);
-typedef void (CALLBACK * FREEADDRINFO) (struct addrinfo FAR *);
-typedef int (CALLBACK * GETNAMEINFO) (const struct sockaddr *, socklen_t,
-    char *, size_t, char *, size_t, int);
-extern GETADDRINFO s_getaddrinfo;
-extern FREEADDRINFO s_freeaddrinfo;
-extern GETNAMEINFO s_getnameinfo;
-#endif /* ! _WIN32_WCE */
-
 #endif /* USE_WIN32 */
 
 int getnameinfo(const struct sockaddr *, socklen_t,
@@ -629,26 +723,76 @@ int getnameinfo(const struct sockaddr *, socklen_t,
 
 /**************************************** prototypes for sthreads.c */
 
+#ifndef USE_FORK
+extern CLI *thread_head;
+#endif
+
+#if OPENSSL_VERSION_NUMBER<0x10100004L
+
+#ifdef USE_OS_THREADS
+
+struct CRYPTO_dynlock_value {
+#ifdef USE_PTHREAD
+    pthread_rwlock_t rwlock;
+#endif
+#ifdef USE_WIN32
+    CRITICAL_SECTION critical_section;
+#endif
+#ifdef DEBUG_LOCKS
+    const char *init_file, *read_lock_file, *write_lock_file,
+        *unlock_file, *destroy_file;
+    int init_line, read_lock_line, write_lock_line, unlock_line, destroy_line;
+#endif
+};
+
+typedef struct CRYPTO_dynlock_value CRYPTO_RWLOCK;
+
+#else /* USE_OS_THREADS */
+
+typedef void CRYPTO_RWLOCK;
+
+#endif /* USE_OS_THREADS */
+
+#endif /* OPENSSL_VERSION_NUMBER<0x10100004L */
+
 typedef enum {
-    CRIT_SESSION, CRIT_ADDR,
-    CRIT_CLIENTS, CRIT_SSL,                 /* client.c */
-    CRIT_INET,                              /* resolver.c */
+    LOCK_THREAD_LIST,                       /* sthreads.c */
+    LOCK_SESSION, LOCK_ADDR,
+    LOCK_CLIENTS, LOCK_SSL,                 /* client.c */
+    LOCK_REF,                               /* options.c */
+    LOCK_INET,                              /* resolver.c */
 #ifndef USE_WIN32
-    CRIT_LIBWRAP,                           /* libwrap.c */
+    LOCK_LIBWRAP,                           /* libwrap.c */
+#endif
+    LOCK_LOG_BUFFER, LOCK_LOG_MODE,         /* log.c */
+    LOCK_LEAK_HASH, LOCK_LEAK_RESULTS,      /* str.c */
+#ifndef OPENSSL_NO_DH
+    LOCK_DH,                                /* ctx.c */
+#endif /* OPENSSL_NO_DH */
+#ifdef USE_WIN32
+    LOCK_WIN_LOG,                           /* ui_win_gui.c */
+#endif
+    LOCK_SECTIONS,                          /* traversing section list */
+    STUNNEL_LOCKS                           /* number of locks */
+} LOCK_TYPE;
+
+extern CRYPTO_RWLOCK *stunnel_locks[STUNNEL_LOCKS];
+
+#if OPENSSL_VERSION_NUMBER<0x10100004L
+/* Emulate the OpenSSL 1.1 locking API for older OpenSSL versions */
+CRYPTO_RWLOCK *CRYPTO_THREAD_lock_new(void);
+int CRYPTO_THREAD_read_lock(CRYPTO_RWLOCK *);
+int CRYPTO_THREAD_write_lock(CRYPTO_RWLOCK *);
+int CRYPTO_THREAD_unlock(CRYPTO_RWLOCK *);
+void CRYPTO_THREAD_lock_free(CRYPTO_RWLOCK *);
+int CRYPTO_atomic_add(int *, int, int *, CRYPTO_RWLOCK *);
 #endif
-    CRIT_LOG, CRIT_LEAK,                    /* log.c */
-#ifndef OPENSSL_NO_DH
-    CRIT_DH,                                /* ctx.c */
-#endif /* OPENSSL_NO_DH */
-    CRIT_SECTIONS                           /* number of critical sections */
-} SECTION_CODE;
 
-void enter_critical_section(SECTION_CODE);
-void leave_critical_section(SECTION_CODE);
 int sthreads_init(void);
 unsigned long stunnel_process_id(void);
 unsigned long stunnel_thread_id(void);
-int create_client(SOCKET, SOCKET, CLI *, void *(*)(void *));
+int create_client(SOCKET, SOCKET, CLI *);
+
 #ifdef USE_UCONTEXT
 typedef struct CONTEXT_STRUCTURE {
     char *stack; /* CPU stack for this thread */
@@ -663,12 +807,15 @@ typedef struct CONTEXT_STRUCTURE {
 extern CONTEXT *ready_head, *ready_tail;
 extern CONTEXT *waiting_head, *waiting_tail;
 #endif
+
 #ifdef _WIN32_WCE
 long _beginthread(void (*)(void *), int, void *);
 void _endthread(void);
 #endif
+
 #ifdef DEBUG_STACK_SIZE
-void stack_info(int);
+void stack_info(size_t, int);
+void ignore_value(void *);
 #endif
 
 /**************************************** prototypes for file.c */
@@ -689,18 +836,18 @@ LPSTR tstr2str(LPCTSTR);
 
 /**************************************** prototypes for libwrap.c */
 
-int libwrap_init();
-void libwrap_auth(CLI *, char *);
+int libwrap_init(void);
+void libwrap_auth(CLI *);
 
 /**************************************** prototypes for tls.c */
 
 extern volatile int tls_initialized;
 
-void tls_init();
-TLS_DATA *tls_alloc(CLI *, TLS_DATA *, char *);
-void tls_cleanup();
+void tls_init(void);
+TLS_DATA *tls_alloc(CLI *, TLS_DATA *, const char *);
+void tls_cleanup(void);
 void tls_set(TLS_DATA *);
-TLS_DATA *tls_get();
+TLS_DATA *tls_get(void);
 
 /**************************************** prototypes for str.c */
 
@@ -712,13 +859,16 @@ struct tls_data_struct {
     size_t alloc_bytes, alloc_blocks;
     CLI *c;
     SERVICE_OPTIONS *opt;
-    char *id;
+    const char *id;
 };
 
-void str_init(TLS_DATA *);
-void str_cleanup(TLS_DATA *);
+void str_init(void);
+void str_thread_init(TLS_DATA *);
+void str_thread_cleanup(TLS_DATA *);
 char *str_dup_debug(const char *, const char *, int);
 #define str_dup(a) str_dup_debug((a), __FILE__, __LINE__)
+char *str_dup_detached_debug(const char *, const char *, int);
+#define str_dup_detached(a) str_dup_detached_debug((a), __FILE__, __LINE__)
 char *str_vprintf(const char *, va_list);
 char *str_printf(const char *, ...)
 #ifdef __GNUC__
@@ -730,19 +880,31 @@ char *str_printf(const char *, ...)
 LPTSTR str_tprintf(LPCTSTR, ...);
 #endif
 
-void str_canary_init();
-void str_stats();
+void str_canary_init(void);
+void str_stats(void);
+
 void *str_alloc_debug(size_t, const char *, int);
 #define str_alloc(a) str_alloc_debug((a), __FILE__, __LINE__)
 void *str_alloc_detached_debug(size_t, const char *, int);
 #define str_alloc_detached(a) str_alloc_detached_debug((a), __FILE__, __LINE__)
+
 void *str_realloc_debug(void *, size_t, const char *, int);
 #define str_realloc(a, b) str_realloc_debug((a), (b), __FILE__, __LINE__)
+void *str_realloc_detached_debug(void *, size_t, const char *, int);
+#define str_realloc_detached(a, b) str_realloc_detached_debug((a), (b), __FILE__, __LINE__)
+
 void str_detach_debug(void *, const char *, int);
 #define str_detach(a) str_detach_debug((a), __FILE__, __LINE__)
+void str_detach_const_debug(const void *, const char *, int);
+#define str_detach_const(a) str_detach_const_debug((a), __FILE__, __LINE__)
+
 void str_free_debug(void *, const char *, int);
 #define str_free(a) str_free_debug((a), __FILE__, __LINE__), (a)=NULL
 #define str_free_expression(a) str_free_debug((a), __FILE__, __LINE__)
+void str_free_const_debug(const void *, const char *, int);
+#define str_free_const(a) str_free_const_debug((a), __FILE__, __LINE__), (a)=NULL
+
+void leak_table_utilization(void);
 
 int safe_memcmp(const void *, const void *, size_t);
 
@@ -757,9 +919,12 @@ void ui_new_log(const char *);
 void message_box(LPCTSTR, const UINT);
 #endif /* USE_WIN32 */
 
-int passwd_cb(char *, int, int, void *);
+int ui_passwd_cb(char *, int, int, void *);
 #ifndef OPENSSL_NO_ENGINE
-int pin_cb(UI *, UI_STRING *);
+int (*ui_get_opener(void)) (UI *);
+int (*ui_get_writer(void)) (UI *, UI_STRING *);
+int (*ui_get_reader(void)) (UI *, UI_STRING *);
+int (*ui_get_closer(void)) (UI *);
 #endif /* !defined(OPENSSL_NO_ENGINE) */
 
 #ifdef ICON_IMAGE
